""" Blueprint for data prediction

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    20.09.2023

This module defines a Flask Blueprint for predicting data.

"""
import os

from flask import Blueprint, request, jsonify
from nova_server.utils.thread_utils import THREADS
from nova_server.utils.job_utils import get_job_id_from_request_form
from nova_server.utils import thread_utils, job_utils, env
from nova_server.utils import log_utils
from nova_server.exec.execution_handler import NovaProcessHandler

process = Blueprint("process", __name__)


@process.route("/process", methods=["POST"])
def predict_thread():
    if request.method == "POST":
        request_form = request.form.to_dict()
        job_id = get_job_id_from_request_form(request_form)
        job_added = job_utils.add_new_job(job_id, request_form=request_form)
        thread = process_data(request_form, job_id)
        thread.start()
        THREADS[job_id] = thread
        data = {"success": str(job_added)}
        return jsonify(data)


@thread_utils.ml_thread_wrapper
def process_data(request_form, job_id):
    #job_id = get_job_id_from_request_form(request_form)
    job_utils.update_status(job_id, job_utils.JobStatus.RUNNING)
    logger = log_utils.get_logger_for_job(job_id)
    logger.info(request_form)

    job = job_utils.get_job(job_id)
    job.execution_handler = NovaProcessHandler(request_form, logger=logger, backend=os.getenv(env.NOVA_SERVER_BACKEND))

    try:
        job.run()
        job_utils.update_status(job_id, job_utils.JobStatus.FINISHED)
    except Exception as e:
        logger.critical(f"Job failed with exception {str(e)}")
        job_utils.update_status(job_id, job_utils.JobStatus.ERROR)



"""Backend for using virtual environments as a backend for nova server

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    18.8.2023

"""

import os
import shutil
import string
import subprocess
import signal
import sys
import time
from threading import Thread
from subprocess import Popen, PIPE
from pathlib import Path
from nova_server.utils import venv_utils as vu
from dotenv import load_dotenv
from logging import Logger

import psutil
import subprocess
import os

class VenvHandler:
    """
    Handles the creation and management of a virtual environment and running scripts within it.

    Args:
        module_dir (Path, optional): The path to the nova server module directory for the environtment.
        logger (Logger, optional): The logger instance for logging.
        log_verbose (bool, optional): If True, log verbose output.

    Attributes:
        venv_dir (Path): The path to the virtual environment.
        log_verbose (bool): If True, log verbose output.
        module_dir (Path): The path to the module directory.
        logger (Logger): The logger instance.

    Example:
        >>> import logging
        >>> load_dotenv("../.env")
        >>> log_dir = Path(os.getenv('NOVA_LOG_DIR', '.'))
        >>> log_file = log_dir / 'test.log'
        >>> logging.basicConfig(filename=log_file, encoding='utf-8', level=logging.DEBUG)
        >>> logger = logging.getLogger('test_logger')
        >>> module_path = Path(os.getenv("NOVA_CML_DIR")) / "test"
        >>> venv_handler = VenvHandler(module_path, logger=logger, log_verbose=True)
        >>> venv_handler.run_python_script_from_file(
        ...     module_path / "test.py",
        ...     script_args=["pos_1", "pos_2"],
        ...     script_kwargs={"-k1": "k1", "--keyword_three": "k3"},
        ... )
    """

    def _reader(self, stream, context=None):
        """
        Reads and logs output from a stream.

        Args:
            stream: The stream to read.
            context (str, optional): The context of the stream (stdout or stderr).
        """
        while True:
            try:
                s = stream.readline()
                printable = set(string.printable)
                s = ''.join(filter(lambda x: x in printable, s))
                if not s or s == "":
                    break
                if self.logger is None:
                    if not self.log_verbose:
                        sys.stderr.write(".")
                    else:
                        sys.stderr.write(s.strip('\n'))
                else:
                    if context == "stderr":
                        self.logger.error(s.strip('\n'))
                    else:
                        self.logger.info(s.strip('\n'))
                sys.stderr.flush()
            except Exception as e:
                continue
               #self.logger.error(e)
        stream.close()

    def _run_cmd(self, cmd: str, wait: bool = True) -> int:
        """
        Executes a command in a subprocess and logs the output.

        Args:
           cmd (str): The command to execute.
           wait (bool, optional): If True, wait for the command to complete.

        Returns:
            int: Return code of the executed command
        """
        # TODO check if CREATE_NEW_PROCESS_GROUP works in unix
        self.current_process = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True, universal_newlines=True)
        t1 = Thread(target=self._reader, args=(self.current_process.stdout, "stdout"))
        t1.start()
        t2 = Thread(target=self._reader, args=(self.current_process.stderr, "stderr"))
        t2.start()
        if wait:
            self.current_process.wait()
            t1.join()
            t2.join()
        return self.current_process.returncode

    def _get_or_create_venv(self):
        """
        Gets or creates a virtual environment and returns its path.

        Returns:
            Path: The path to the virtual environment.
        """
        venv_dir = vu.venv_dir_from_mod(self.module_dir)
        existed = True
        if not venv_dir.is_dir():
            existed = False
            try:
                run_cmd = f"{sys.executable} -m venv {venv_dir}"
                self._run_cmd(run_cmd)
            except Exception as e:
                shutil.rmtree(venv_dir)
                raise e
        return venv_dir, existed

    def _upgrade_pip(self):
        """
        Upgrades the `pip` package within the virtual environment.
        """

        args = ["install"]
        if not self.log_verbose:
            args.append("-q")

        run_cmd = vu.get_module_run_cmd(
            self.venv_dir, "pip", args=args, kwargs={"--upgrade": "pip"}
        )
        self._run_cmd(run_cmd)

    def _install_requirements(self):
        """
        Installs requirements from a `requirements.txt` file within the virtual environment.
        """
        # pip
        self._upgrade_pip()

        # hcai-nova-utils
        #self._install_nova_utils()

        # requirements.txt
        req_txt = self.module_dir / "requirements.txt"
        if not req_txt.is_file():
            return
        else:
            # TODO: replace with custom import utils for better parsing
            # requirements.txt

            args = ["install"]
            if not self.log_verbose:
                args.append("-q")
            run_cmd = vu.get_module_run_cmd(
                self.venv_dir,
                "pip",
                args=args,
                kwargs={"-r": str(req_txt.resolve())},
            )
            self._run_cmd(run_cmd)

    def _install_nova_utils(self):
        import toml

        toml_file = Path(__file__) / '..' / '..' / '..' / 'pyproject.toml'
        with open(toml_file.resolve(), 'r') as f:
            config = toml.load(f)
            dependencies = config['project']['dependencies']
            nu_dependency = list(filter(lambda x: x.startswith('hcai-nova-utils'), dependencies))
            if len(nu_dependency) == 1:
                args = ["install"]
                if not self.log_verbose:
                    args.append("-q")
                args.append(nu_dependency[0])

                run_cmd = vu.get_module_run_cmd(
                    self.venv_dir,
                    "pip",
                    args=args
                )
                self._run_cmd(run_cmd)
            else:
                raise ValueError(f'Could not find unique version for hcai-nova-utils in {dependencies}')

    def __init__(
        self, module_dir: Path = None, logger: Logger = None, log_verbose: bool = False, force_requirements: bool = False
    ):
        """
        Initializes the VenvHandler instance.

        Args:
            module_dir (Path, optional): The path to the module directory.
            logger (Logger, optional): The logger instance for logging.
            log_verbose (bool, optional): If True, log verbose output.
            force_requirements(bool, optional): If True, requirement.txt will always be installed. If false skipping requirements installation if venv already exist.
        """
        self.venv_dir = None
        self.current_process = None
        self.log_verbose = log_verbose
        self.module_dir = module_dir
        self.logger = logger if logger is not None else Logger(__name__)
        self.force_requirements = force_requirements
        if module_dir is not None:
            self.init_venv()

    def init_venv(self):
        """
        Initializes the virtual environment and installs requirements.
        """
        self.logger.info(f"Initializing venv")
        venv_dir, existed = self._get_or_create_venv()
        self.venv_dir = venv_dir
        if not existed or self.force_requirements:
            self._install_requirements()
        self.logger.info(f"Venv {self.venv_dir} initialized")

    def run_python_script_from_file(
        self, script_fp: Path, script_args: list = None, script_kwargs: dict = None
    ):
        """
        Runs a Python script within the virtual environment.

        Args:
            script_fp (Path): The path to the script to run.
            script_args (list, optional): List of positional arguments to pass to the script.
            script_kwargs (dict, optional): Dictionary of keyword arguments to pass to the script.

        Raises:
            ValueError: If the virtual environment has not been initialized. Call `init_venv()` first.
            subprocess.CalledProcessError: If the executed command exits with a return value other than 0
        """
        if self.venv_dir is None:
            raise ValueError(
                "Virtual environment has not been initialized. Call <init_venv()> first."
            )
        run_cmd = vu.get_python_script_run_cmd(
            self.venv_dir, script_fp, script_args, script_kwargs
        )
        return_code = self._run_cmd(run_cmd)
        if not return_code == 0:
            raise subprocess.CalledProcessError(returncode=return_code, cmd=run_cmd)

    def run_shell_script(self, script: str, script_args: list = None, script_kwargs: dict = None):
        """
        Runs a command in the respective os shell with an activated virtual environment


        Args:
            script (Path): The path to the script to run.
            script_args (list, optional): List of positional arguments to pass to the script.
            script_kwargs (dict, optional): Dictionary of keyword arguments to pass to the script.

        Raises:
            ValueError: If the virtual environment has not been initialized. Call `init_venv()` first.
            subprocess.CalledProcessError: If the executed command exits with a return value other than 0
        """
        if self.venv_dir is None:
            raise ValueError(
                "Virtual environment has not been initialized. Call <init_venv()> first."
            )
        run_cmd = vu.get_shell_script_run_cmd(
            self.venv_dir, script, script_args, script_kwargs
        )
        return_code = self._run_cmd(run_cmd)
        if not return_code == 0:
            raise subprocess.CalledProcessError(returncode=return_code, cmd=run_cmd)

    def kill(self):
        '''Kills parent and children processess'''
        parent = psutil.Process(self.current_process.pid)
        # kill all the child processes
        for child in parent.children(recursive=True):
            try:
                self.logger.info(f'Killing process {child}')
                child.kill()
                # kill the parent process
                self.logger.info(f'Killing process {parent}')
                parent.kill()
            except psutil.NoSuchProcess:
                continue

if __name__ == "__main__":
    import logging

    load_dotenv("../.env")

    log_dir = Path(os.getenv("NOVA_LOG_DIR", "."))
    log_file = log_dir / "test.log"
    logging.basicConfig(filename=log_file, encoding="utf-8", level=logging.DEBUG)
    logger = logging.getLogger("test_logger")

    module_path = Path(os.getenv("NOVA_CML_DIR")) / "test"
    venv_handler = VenvHandler(module_path, logger=logger, log_verbose=True)
    venv_handler.run_python_script_from_file(
        module_path / "test.py",
        script_args=["pos_1", "pos_2"],
        script_kwargs={"-k1": "k1", "--keyword_three": "k3"},
    )



""" Blueprint for querying server capabilities

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    15.11.2023

This module defines a Flask Blueprint for uploading files to the server.
Returns the filepath where the data is stored on the server
"""
import os
import json
from flask import Blueprint, request, redirect, url_for
from pathlib import Path
import tempfile
import random
import string

upload = Blueprint("upload", __name__)
@upload.route("/upload", methods=["POST"])
def upload_file():
    """
    Upload a file to the server for further processing

    Returns:
        str: A relative filepath to the uploaded file

    Example:

    """
    if request.method == "POST":
        file = request.files['file']

        upload_dir = Path(tempfile.tempdir)
        upload_dir.mkdir(parents=False, exist_ok=True)

        fn = ''.join(random.choices(string.ascii_lowercase, k=6)) + '.'+file.filename.split('.')[-1]
        fp = upload_dir / fn
        file.save(fp)
        rel_path = Path(fp.resolve()).relative_to(Path(tempfile.gettempdir()).resolve())
        return str(rel_path)





""" Blueprint for querying server capabilities

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    14.09.2023

This module defines a Flask Blueprint for querying the server for its capabilities.
Returns a list of all available server-module trainer and chains in json format.

"""
import os
import json
from flask import Blueprint, request
from nova_server.utils import env
from nova_utils.utils.json_utils import ChainEncoder, TrainerEncoder
from nova_utils.utils.ssi_xml_utils import Chain, Trainer
from nova_utils.utils.request_utils import parse_cml_filter
from pathlib import Path
from glob import glob

cml_info = Blueprint("cml_info", __name__)

def filter(trainer_io_list: list, filter_list: list) -> bool:

    if len(trainer_io_list) != len(filter_list):
        return False

    for f in filter_list:
        super_type_f, sub_type_f, specific_type_f = parse_cml_filter(f)
        for t in trainer_io_list:
            super_type_t, sub_type_t, specific_type_t = parse_cml_filter(t.io_data)

            check_super_type =  super_type_f == super_type_t
            check_sub_type =  (sub_type_f is not None and sub_type_f == sub_type_t) or sub_type_f is None
            check_specifc_type =  (specific_type_f is not None and specific_type_f == specific_type_t) or specific_type_f is None
            if not (check_super_type and check_sub_type and check_specifc_type):
                return False
    return True

@cml_info.route("/cml_info", methods=["POST"])
def info():
    """
    Query the server to return information about all

    This route allows querying the server for all available nova-server modules

    Returns:
        dict: A JSON response containing the status of the requested job.

    Example:

    """
    if request.method == "POST":

        cml_dir = os.getenv(env.NOVA_SERVER_CML_DIR)

        if not cml_dir:
            return None

        cml_dir = str(Path(cml_dir).resolve())

        trainer_files = glob(cml_dir + '/**/*.trainer', recursive = True)
        chain_files = glob(cml_dir + '/**/*.chain', recursive = True)

        trainer_ok = {}
        trainer_faulty = {}
        chains_ok = {}
        chains_faulty = {}

        request_form = request.form.to_dict()
        input_filter = json.loads(request_form.get('input_filter', '[]'))
        output_filter = json.loads(request_form.get('output_filter', '[]'))


        for tf in trainer_files:
            t = Trainer()
            rtf = str(Path(tf).relative_to(cml_dir))
            try:
                t.load_from_file(tf)
                inputs_ok = filter([x for x in t.meta_io if x.io_type=='input'], input_filter) if input_filter else True
                outputs_ok = filter([x for x in t.meta_io if x.io_type=='output'], output_filter) if output_filter else True
                if not (inputs_ok and outputs_ok):
                    continue
                trainer_ok[rtf] = json.dumps(t, cls=TrainerEncoder)
            except Exception as e:
                trainer_faulty[rtf] = str(e)

        for cf in chain_files:
            c = Chain()
            rcf = str(Path(cf).relative_to(cml_dir))
            try:
                c.load_from_file(cf)
                chains_ok[rcf] = json.dumps(c, cls=ChainEncoder)
            except Exception as e:
                chains_faulty[rcf] = str(e)


        return {
            'chains_ok' : chains_ok,
            'chains_faulty' : chains_faulty,
            'trainer_ok' : trainer_ok,
            'trainer_faulty' : trainer_faulty
        }






""" Class definition for th execution handler. The execution handler combines uses data from the request form to execute
the correct command using the specified backends.

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023
"""
import importlib
import os
import re
from logging import Logger
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from pathlib import PureWindowsPath
from nova_server.utils import env


class Action(Enum):
    PROCESS = "nu-process"
    TRAIN = "nu-train"

class Backend(Enum):
    DEBUG = 'debug'
    VENV = 'venv'


class ExecutionHandler(ABC):
    @property
    def script_arguments(self):
        return self._script_arguments

    @script_arguments.setter
    def script_arguments(self, value):
        # convert dictionary keys from camelCase to snake_case
        self._script_arguments = {
            "--" + re.sub(r"(?<!^)(?=[A-Z])(?<![A-Z])", "_", k).lower(): v
            for k, v in value.items()
        }

    def __init__(
        self, request_form: dict, backend: str = 'venv', logger: Logger = None
    ):
        self.backend = Backend(backend)
        self.script_arguments = request_form
        self.logger = logger
        self.backend_handler = None

    def _nova_server_env_to_arg(self):
        arg_vars = {}
        prefix = "NOVA_SERVER_"

        # Select data folder, where dataset exists if multiple folders are passed
        dd = os.getenv(env.NOVA_SERVER_DATA_DIR)
        ds = self.script_arguments.get('--dataset')
        if ds is not None:
            if ';' in dd:
                for dir in dd.split(';'):
                    dataset_dir = (Path(dir) / ds)
                    if dataset_dir.is_dir():
                        os.environ[env.NOVA_SERVER_DATA_DIR] = str(dataset_dir.resolve())
                        break

        # set other vars
        env_vars = [
                env.NOVA_SERVER_CML_DIR,
                env.NOVA_SERVER_DATA_DIR,
                env.NOVA_SERVER_LOG_DIR,
                env.NOVA_SERVER_CACHE_DIR,
                env.NOVA_SERVER_TMP_DIR,
        ]

        for var in env_vars:
            k = "--" + var[len(prefix) :].lower()
            v = os.getenv(var)
            arg_vars[k] = v
        return arg_vars

    def run(self):

        # Run with selected backend
        if self.backend == Backend.DEBUG:
            from importlib.machinery import SourceFileLoader
            import nova_utils as nu
            from importlib.metadata import entry_points
            ep = [x for x in entry_points().get('console_scripts') if x.name == self.run_script ][0]
            run_module = importlib.import_module(ep.module)
            main = getattr(run_module, 'main')

            # Add dotenv variables to arguments for script
            self._script_arguments |= self._nova_server_env_to_arg()
            self._script_arguments.setdefault('--shared_dir', os.getenv(env.NOVA_SERVER_TMP_DIR))

            args = []
            for k,v in self._script_arguments.items():
                args.append(k)
                args.append(v)

            main(args)

        elif self.backend == Backend.VENV:
            from nova_server.backend import virtual_environment as backend

            # Setup virtual environment
            cml_dir = os.getenv(env.NOVA_SERVER_CML_DIR)
            if cml_dir is None:
                raise ValueError(f"NOVA_CML_DIR not set")

            module_dir = Path(cml_dir) / self.module_name
            if not module_dir.is_dir():
                raise NotADirectoryError(
                    f"NOVA_CML_DIR {module_dir} is not a valid directory"
                )

            self.backend_handler = backend.VenvHandler(
                module_dir, logger=self.logger, log_verbose=True
            )

            # Add dotenv variables to arguments for script
            self._script_arguments |= self._nova_server_env_to_arg()
            self._script_arguments.setdefault('--shared_dir', os.getenv(env.NOVA_SERVER_TMP_DIR))


            self.backend_handler.run_shell_script(
                self.run_script,
                script_kwargs=self._script_arguments,
            )

        else:
            raise ValueError(f"Unknown backend {self.backend}")

    def cancel(self):
        if self.backend == Backend.VENV:
            self.backend_handler.kill()

    @property
    @abstractmethod
    def run_script(self):
        pass

    @property
    @abstractmethod
    def module_name(self):
        pass


class NovaProcessHandler(ExecutionHandler):
    @property
    def module_name(self):
        tfp = self.script_arguments.get("--trainer_file_path")
        if tfp is None:
            raise ValueError("trainerFilePath not specified in request.")
        else:
            return PureWindowsPath(tfp).parent

    @property
    def run_script(self):
        return self.action.value

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.action = Action.PROCESS

class NovaTrainHandler(ExecutionHandler):
    @property
    def module_name(self):
        tfp = self.script_arguments.get("--trainer_file_path")
        if tfp is None:
            raise ValueError("trainerFilePath not specified in request.")
        else:
            return PureWindowsPath(tfp).parent

    @property
    def run_script(self):
        return self.action.value

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.action = Action.TRAIN



""" Blueprint for retrieving a job's log file

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    27.10.2023

This module defines a Flask Blueprint for retrieving a job's log file.

"""
import flask
from flask import Blueprint, request, abort
from nova_server.utils.job_utils import get_job_id_from_request_form
from nova_server.utils import env
from pathlib import Path
import tempfile
import shutil
fetch_result = Blueprint("fetch_result", __name__)

import zipfile
import os
from nova_utils.utils.string_utils import string_to_bool
from flask import send_file, current_app
from mimetypes import guess_type

def supported_file(x: Path):
    file_names = [
    ]
    file_starts_with = [
        '.'
    ]
    file_ext = [

    ]
    return not (x.name in file_names or any([x.name.startswith(s) for s in file_starts_with]) or x.suffix in file_ext)

#https://stackoverflow.com/questions/24612366/delete-an-uploaded-file-after-downloading-it-from-flask
#https://stackoverflow.com/questions/40853201/remove-file-after-flask-serves-it/40854330#40854330
def download_and_remove_dir(job_dir: Path, delete_after_download=False):

    to_del = []
    if delete_after_download:
        to_del.append(job_dir)

    files = list(filter(lambda x: supported_file(x) and x.is_file(), job_dir.rglob('*')))

    download_fp = None
    if len(files) > 1:
        zip_fp = Path(tempfile.gettempdir()) / (job_dir.name + '.zip')
        zipfolder = zipfile.ZipFile(zip_fp, 'w', compression=zipfile.ZIP_STORED)
        for file in files:
            zipfolder.write(file, arcname=file.relative_to(job_dir.parent))
        zipfolder.close()
        download_fp = zip_fp
        to_del.append(download_fp)
    elif len(files) == 1:
        download_fp = files[0]
    else:
        raise FileNotFoundError(f'{files}')

    def generate():
        with open(download_fp, mode="rb") as f:
            while chunk := f.read(1024):
                yield chunk

        for d in to_del:
            d: Path
            print(f'Deleting {d}')
            if d.is_file():
                os.remove(d)
            if d.is_dir():
                shutil.rmtree(d)



    mt = guess_type(download_fp)[0]
    r = current_app.response_class(generate(), mimetype=mt)
    r.headers.set('Content-Disposition', 'attachment', filename=download_fp.name)
    return r


@fetch_result.route("/fetch_result", methods=["POST"])
def fetch_thread():
    """
    Retrieve the results of a specific job.

    This route allows retrieving the data file  for a job by providing the job's unique identifier in the request.

    Returns:
        dict: Data object for the respective job. 404 if not data has been found

    Example:
        >>> POST /log
        >>> {"job_id": "12345"}
        {"message": "Log file content here..."}
    """
    if request.method == "POST":
        request_form = request.form.to_dict()
        job_id = get_job_id_from_request_form(request_form)
        delete_after_download = request_form.get('delete_after_download', False)
        if isinstance(delete_after_download, str):
            delete_after_download = string_to_bool(delete_after_download)

        shared_dir = os.getenv(env.NOVA_SERVER_TMP_DIR)
        job_dir = Path(shared_dir) / job_id

        if not job_dir.exists():
            raise FileNotFoundError(f'No data for job id {job_id} found on server.')

        else:
            return download_and_remove_dir(job_dir, delete_after_download)


